"""
An empty pert file.
"""


def pert(t, system):
    """
    Perturbation function called at each step.

    The function is named "pert" and takes two arguments:
    ``t`` for simulation time, and ``system`` for the system object.

    """
    pass
